import flask

from ...contracts_generated import (
    AbstraLibApiEditorLintersCheckResponse,
    AbstraLibApiEditorLintersFixResponse,
    AbstraLibApiEditorLintersRule,
)
from ...linter.rules import rules
from ...usage import usage


def check_linters() -> AbstraLibApiEditorLintersCheckResponse:
    return [AbstraLibApiEditorLintersRule.from_dict(rule.to_dict()) for rule in rules]


def fix_linter(rule_name: str, fix_name: str):
    for rule in rules:
        if rule.name == rule_name:
            for issue in rule.find_issues():
                for fix in issue.fixes:
                    if fix.name == fix_name:
                        fix.fix()
                        return True
    raise Exception(f"Could not find fix {fix_name} for rule {rule_name}")


def get_editor_bp():
    bp = flask.Blueprint("editor_linters", __name__)

    # 1s pooling in this route
    @bp.get("/check")
    def _check_linters():
        return check_linters()

    @bp.post("/fix/<rule_name>/<fix_name>")
    @usage
    def _fix_linter(rule_name: str, fix_name: str):
        fix_linter(rule_name, fix_name)
        return AbstraLibApiEditorLintersFixResponse(success=True).to_dict()

    return bp



from .CnpjInput import CnpjInput
from .NpsInput import NpsInput
from .VideoInput import VideoInput
from .CustomInput import CustomInput
from .PandasRowSelectionInput import PandasRowSelectionInput
from .NumberInput import NumberInput
from .TagInput import TagInput
from .TextareaInput import TextareaInput
from .ImageInput import ImageInput
from .RatingInput import RatingInput
from .CurrencyInput import CurrencyInput
from .ListInput import ListInput
from .TextOutput import TextOutput
from .ProgressOutput import ProgressOutput
from .CameraInput import CameraInput
from .CpfInput import CpfInput
from .PandasOutput import PandasOutput
from .EmailInput import EmailInput
from .CheckboxInput import CheckboxInput
from .FileInput import FileInput
from .MultipleChoiceInput import MultipleChoiceInput
from .PhoneInput import PhoneInput
from .DateInput import DateInput
from .PlotlyOutput import PlotlyOutput
from .CardsInput import CardsInput
from .HtmlOutput import HtmlOutput
from .CodeInput import CodeInput
from .IframeOutput import IframeOutput
from .LatexOutput import LatexOutput
from .ImageOutput import ImageOutput
from .FileOutput import FileOutput
from .LinkOutput import LinkOutput
from .ChecklistInput import ChecklistInput
from .TimeInput import TimeInput
from .PasswordInput import PasswordInput
from .RichTextInput import RichTextInput
from .ToggleInput import ToggleInput
from .DropdownInput import DropdownInput
from .TextInput import TextInput
from .MarkdownOutput import MarkdownOutput
from .NumberSliderInput import NumberSliderInput



from io import BufferedReader, TextIOWrapper
from pathlib import Path
from typing import List, Union, Optional
from ..apis import upload_file
from ..response_types import FileResponse
from ..widget_base import Input, MultipleHandler


class FileInput(Input):
    type = 'file-input'
    multiple: bool = False
    multiple_handler: MultipleHandler

    def __init__(self, key: str, label: str, **kwargs):
        super().__init__(key)
        self.set_props(dict(label=label, **kwargs))

    def set_props(self, props):
        self.label = props.get('label', 'Label')
        self.required = props.get('required', True)
        self.hint = props.get('hint', None)
        self.full_width = props.get('full_width', False)
        self.disabled = props.get('disabled', False)
        self.max_file_size = props.get('max_file_size', None)
        self.multiple = props.get('multiple', False)
        self.value = props.get('initial_value', self.empty_value)
        self.multiple_handler = MultipleHandler(self.multiple, required=
            self.required)

    def render(self, ctx: dict):
        return {'type': self.type, 'key': self.key, 'hint': self.hint,
            'label': self.label, 'value': self.serialize_value(),
            'required': self.required, 'multiple': self.multiple,
            'fullWidth': self.full_width, 'disabled': self.disabled,
            'maxFileSize': self.max_file_size, 'errors': self.errors}

    @staticmethod
    def __get_file_uri(value: Union[FileResponse, str, BufferedReader,
        TextIOWrapper]) ->str:
        if isinstance(value, str):
            if Path(value).is_file():
                return upload_file(open(value))
            else:
                return value
        if isinstance(value, FileResponse):
            return value.url
        if isinstance(value, BufferedReader):
            return upload_file(value)
        if isinstance(value, TextIOWrapper):
            return upload_file(value)
        return ''

    def serialize_value(self) ->List[str]:
        values_list = self.multiple_handler.value_to_list(self.value)
        return [FileInput.__get_file_uri(item) for item in values_list]

    def parse_value(self, value: Optional[List[str]]) ->Union[FileResponse,
        List[FileResponse], None]:
        file_responses = [FileResponse(item) for item in value or []]
        return self.multiple_handler.value_to_list_or_value(file_responses)



from typing import List, Union

from ...repositories.project.project import (
    ActionStage,
    FormStage,
    HookStage,
    JobStage,
    ProjectRepository,
    ScriptStage,
)
from ...templates import new_form_code, new_hook_code, new_job_code, new_script_code
from ..linter import LinterFix, LinterIssue, LinterRule


class AddEntrypoint(LinterFix):
    label = "Add entrypoint"
    description = "Creates the .py file for the entrypoint"
    stage: Union[FormStage, HookStage, JobStage, ScriptStage]

    def __init__(
        self, stage: Union[FormStage, HookStage, JobStage, ScriptStage]
    ) -> None:
        self.stage = stage

    def make_label(self):
        return f"Create {self.stage.file}"

    def fix(self):
        if isinstance(self.stage, FormStage):
            self.stage.file_path.write_text(new_form_code)
        elif isinstance(self.stage, HookStage):
            self.stage.file_path.write_text(new_hook_code)
        elif isinstance(self.stage, JobStage):
            self.stage.file_path.write_text(new_job_code)
        elif isinstance(self.stage, ScriptStage):
            self.stage.file_path.write_text(new_script_code)
        else:
            raise Exception(f"Unknown stage: {self.stage}")

    @property
    def name(self):
        return (
            f"{self.__class__.__name__}:{self.stage.__class__.__name__}:{self.stage.id}"
        )


class NoEntrypointFound(LinterIssue):
    def __init__(self, stage: ActionStage) -> None:
        self.label = f"The {stage.type_name} entitled {stage.title} points to a non-existent file: {stage.file}"
        self.fixes = [AddEntrypoint(stage)]


class MissingEntrypoint(LinterRule):
    label = "Pointed files should exist"
    type = "bug"

    def find_issues(self) -> List[LinterIssue]:
        project = ProjectRepository.load()
        issues = []
        for form in project.forms:
            if not form.file_path.exists():
                issues.append(NoEntrypointFound(form))

        for hook in project.hooks:
            if not hook.file_path.exists():
                issues.append(NoEntrypointFound(hook))

        for job in project.jobs:
            if not job.file_path.exists():
                issues.append(NoEntrypointFound(job))

        for script in project.scripts:
            if not script.file_path.exists():
                issues.append(NoEntrypointFound(script))

        return issues



from dataclasses import dataclass
from functools import wraps
from typing import Callable, Optional, Union

import flask

from ...contracts_generated import (
    AbstraLibApiPlayerUserNavigationGuard as NavigationGuard,
)
from ...contracts_generated import (
    CommonUserRolesItem,
)
from ...jwt_auth import UserClaims
from ...repositories.project.project import (
    AccessSettings,
    Project,
    ProjectRepository,
)
from ...repositories.users import UsersRepository


def default_auth_decoder(auth_header: str) -> Optional[UserClaims]:
    return UserClaims.from_jwt(auth_header.split(" ")[1])


@dataclass
class StageIdSelector:
    value: str


@dataclass
class PathArgSelector:
    value: str


@dataclass
class IdArgSelector:
    value: str


Selector = Union[StageIdSelector, PathArgSelector, IdArgSelector]


class Guard:
    repository: UsersRepository
    project: Project
    auth_decoder: Callable[[str], Optional[UserClaims]]

    def __init__(
        self,
        repository: UsersRepository,
        auth_decoder: Callable[[str], Optional[UserClaims]] = default_auth_decoder,
        project: Optional[Project] = None,
    ) -> None:
        self.repository = repository
        self.auth_decoder = auth_decoder
        if project is None:
            self.project = ProjectRepository.load()
        else:
            self.project = project

    def __get_access_settings(self, selector: Selector, **kwargs):
        if isinstance(selector, IdArgSelector):
            stage_id = str(kwargs.get(selector.value))
            return self.project.get_access_control_by_stage_id(stage_id)

        elif isinstance(selector, StageIdSelector):
            return self.project.get_access_control_by_stage_id(selector.value)

        ##path arg is default
        path = str(kwargs.get(selector.value))
        return self.project.get_access_control_by_stage_path(path)

    def __allow(self, access_setting: AccessSettings) -> NavigationGuard:
        ##public access
        if access_setting.is_public:
            return NavigationGuard("ALLOWED")

        auth = flask.request.headers.get("Authorization")
        if auth is None:
            return NavigationGuard("UNAUTHORIZED")

        claims = self.auth_decoder(auth)
        if claims is None or claims.email is None:
            return NavigationGuard("UNAUTHORIZED")

        user = self.repository.get_user(claims.email)
        if user is None:
            return NavigationGuard("FORBIDEN")

        ##protected access
        if len(access_setting.required_roles) == 0:
            return NavigationGuard("ALLOWED")

        if len(set(user.roles).intersection(set(access_setting.required_roles))) > 0:
            return NavigationGuard("ALLOWED")

        ##private access
        return NavigationGuard("FORBIDEN")

    def allow(self, path: str) -> NavigationGuard:
        access_settings = self.project.get_access_control_by_stage_path(path)
        ##404 allow this to be handled by the frontend
        if access_settings is None:
            return NavigationGuard("ALLOWED")

        return self.__allow(access_settings)

    def requires(self, role: CommonUserRolesItem):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                auth = flask.request.headers.get("Authorization")
                if auth is None:
                    flask.abort(401)

                claims = self.auth_decoder(auth)
                if claims is None or claims.email is None:
                    flask.abort(401)

                user_roles = self.repository.get_user_roles(claims.email)
                if len(user_roles) == 0 or role not in user_roles:
                    flask.abort(403)

                return func(*args, **kwargs)

            return wrapper

        return decorator

    def by(self, selector: Selector):
        def decorator(func):
            @wraps(func)
            def wrapper(*args, **kwargs):
                access_settings = self.__get_access_settings(selector, **kwargs)
                if access_settings is None:
                    flask.abort(401)

                status = self.__allow(access_settings).status

                if status == "ALLOWED":
                    return func(*args, **kwargs)

                elif status == "UNAUTHORIZED":
                    flask.abort(401)

                ##status == "FORBIDEN"
                flask.abort(403)

            return wrapper

        return decorator
